@@ -1,18 +1,18 @@
-"""
+"""Exception classes thrown by filesystem operations.
 
-Defines the Exception classes thrown by PyFilesystem objects.
+Errors relating to the underlying filesystem are translated in
+to one of the following exceptions.
 
-Errors relating to the underlying filesystem are translated in to one of
-the following exceptions.
-
-All Exception classes are derived from :class:`~fs.errors.FSError` which
-may be used as a catch-all filesystem exception.
+All Exception classes are derived from `~fs.errors.FSError`
+which may be used as a catch-all filesystem exception.
 
 """
 
 from __future__ import unicode_literals
 from __future__ import print_function
 
+import functools
+
 import six
 from six import text_type
 
@@ -31,6 +31,7 @@ __all__ = [
     'InvalidCharsInPath',
     'InvalidPath',
     'MissingInfoNamespace',
+    'NoSysPath',
     'NoURL',
     'OperationFailed',
     'OperationTimeout',
@@ -42,12 +43,14 @@ __all__ = [
     'ResourceInvalid',
     'ResourceLocked',
     'ResourceNotFound',
+    'ResourceReadOnly',
     'Unsupported',
 ]
 
 
 class MissingInfoNamespace(AttributeError):
-    """Raised when an expected namespace was missing."""
+    """An expected namespace is missing.
+    """
 
     def __init__(self, namespace):
         msg = "namespace '{}' is required for this attribute"
@@ -58,7 +61,8 @@ class MissingInfoNamespace(AttributeError):
 
 @six.python_2_unicode_compatible
 class FSError(Exception):
-    """Base exception class for the FS module."""
+    """Base exception for the `fs` module.
+    """
 
     default_message = "Unspecified error"
 
@@ -67,7 +71,8 @@ class FSError(Exception):
         super(FSError, self).__init__()
 
     def __str__(self):
-        """The error message."""
+        """Return the error message.
+        """
         msg = self._msg.format(**self.__dict__)
         return msg
 
@@ -77,19 +82,38 @@ class FSError(Exception):
 
 
 class FilesystemClosed(FSError):
-    """An exception thrown when attempting to use a closed filesystem."""
+    """Attempt to use a closed filesystem.
+    """
 
     default_message = "attempt to use closed filesystem"
 
 
 class CreateFailed(FSError):
-    """An exception thrown when a FS could not be created."""
+    """Filesystem could not be created.
+    """
 
-    default_message = "unable to create filesystem"
+    default_message = "unable to create filesystem, {details}"
 
+    @classmethod
+    def catch_all(cls, func):
+        @functools.wraps(func)
+        def new_func(*args, **kwargs):
+            try:
+                return func(*args, **kwargs)
+            except cls:
+                raise
+            except Exception as e:
+                raise cls(exc=e)
+        return new_func
+
+    def __init__(self, msg=None, exc=None):
+        self._msg = msg or self.default_message
+        self.details = '' if exc is None else text_type(exc)
+        self.exc = exc
 
 class PathError(FSError):
-    """Exception for errors to do with a path string."""
+    """Base exception for errors to do with a path string.
+    """
 
     default_message = "path '{path}' is invalid"
 
@@ -99,13 +123,15 @@ class PathError(FSError):
 
 
 class NoSysPath(PathError):
-    """Exception raised when there is no sys path."""
+    """The filesystem does not provide *sys paths* to the resource.
+    """
 
     default_message = "path '{path}' does not map to the local filesystem"
 
 
 class NoURL(PathError):
-    """Raised when there is no URL for a given path."""
+    """The filesystem does not provide an URL for the resource.
+    """
 
     default_message = "path '{path}' has no '{purpose}' URL"
 
@@ -115,20 +141,22 @@ class NoURL(PathError):
 
 
 class InvalidPath(PathError):
-    """Base exception for fs paths that can't be mapped on to the
-    underlaying filesystem."""
+    """Path can't be mapped on to the underlaying filesystem.
+    """
 
     default_message = "path '{path}' is invalid on this filesystem "
 
 
 class InvalidCharsInPath(InvalidPath):
-    """The path contains characters that are invalid on this filesystem."""
+    """Path contains characters that are invalid on this filesystem.
+    """
 
     default_message = "path '{path}' contains invalid characters"
 
 
 class OperationFailed(FSError):
-    """Base exception class for errors associated with a specific operation."""
+    """A specific operation failed.
+    """
 
     default_message = "operation failed, {details}"
 
@@ -141,43 +169,50 @@ class OperationFailed(FSError):
 
 
 class Unsupported(OperationFailed):
-    """Exception raised for operations that are not supported by the FS."""
+    """Operation not supported by the filesystem.
+    """
 
     default_message = "not supported"
 
 
 class RemoteConnectionError(OperationFailed):
-    """Exception raised when operations encounter remote connection trouble."""
+    """Operations encountered remote connection trouble.
+    """
 
     default_message = "remote connection error"
 
 
 class InsufficientStorage(OperationFailed):
-    """Exception raised when operations encounter storage space trouble."""
+    """Storage is insufficient for requested operation.
+    """
 
     default_message = "insufficient storage space"
 
 
 class PermissionDenied(OperationFailed):
-    """Permissions error."""
+    """Not enough permissions.
+    """
 
     default_message = "permission denied"
 
 
 class OperationTimeout(OperationFailed):
-    """Filesystem took too long."""
+    """Filesystem took too long.
+    """
 
     default_message = "operation timed out"
 
 
 class RemoveRootError(OperationFailed):
-    """Attempt to remove the root directory."""
+    """Attempt to remove the root directory.
+    """
 
     default_message = "root directory may not be removed"
 
 
 class ResourceError(FSError):
-    """Base exception class for error associated with a specific resource."""
+    """Base exception class for error associated with a specific resource.
+    """
 
     default_message = "failed on path {path}"
 
@@ -188,77 +223,84 @@ class ResourceError(FSError):
 
 
 class ResourceNotFound(ResourceError):
-    """Exception raised when a required resource is not found."""
+    """Required resource not found.
+    """
 
     default_message = "resource '{path}' not found"
 
 
 class ResourceInvalid(ResourceError):
-    """Exception raised when a resource is the wrong type."""
+    """Resource has the wrong type.
+    """
 
     default_message = "resource '{path}' is invalid for this operation"
 
 
 class FileExists(ResourceError):
-    """Exception raises when opening a file in exclusive mode."""
+    """File already exists.
+    """
 
     default_message = "resource '{path}' exists"
 
 
 class FileExpected(ResourceInvalid):
-    """Exception raises when a file was expected."""
+    """Operation only works on files.
+    """
 
     default_message = "path '{path}' should be a file"
 
 
 class DirectoryExpected(ResourceInvalid):
-    """Exception raises when a directory was expected."""
+    """Operation only works on directories.
+    """
 
     default_message = "path '{path}' should be a directory"
 
 
 class DestinationExists(ResourceError):
-    """Exception raised when a target destination already exists."""
+    """Target destination already exists.
+    """
 
     default_message = "destination '{path}' exists"
 
 
 class DirectoryExists(ResourceError):
-    """Exception raised when trying to make a directory that already
-    exists."""
+    """Directory already exists.
+    """
 
     default_message = "directory '{path}' exists"
 
 
 class DirectoryNotEmpty(ResourceError):
-    """Exception raised when a directory to be removed is not empty."""
+    """Attempt to remove a non-empty directory.
+    """
 
     default_message = "directory '{path}' is not empty"
 
 
 class ResourceLocked(ResourceError):
-    """Exception raised when a resource can't be used because it is locked."""
+    """Attempt to use a locked resource.
+    """
 
     default_message = "resource '{path}' is locked"
 
 
 class ResourceReadOnly(ResourceError):
-    """Raised when attempting to modify a read only resource."""
+    """Attempting to modify a read-only resource.
+    """
 
     default_message = "resource '{path}' is read only"
 
 
 class IllegalBackReference(ValueError):
-    """
-    Exception raised when too many backrefs exist in a path.
+    """Too many backrefs exist in a path.
 
     This error will occur if the back references in a path would be
     outside of the root. For example, ``"/foo/../../"``, contains two back
     references which would reference a directory above the root.
 
-    .. note::
-
-        This exception is a subclass of ``ValueError`` as it is not
+    Note:
+        This exception is a subclass of `ValueError` as it is not
         strictly speaking an issue with a filesystem or resource.
 
     """
