@@ -1,3 +1,6 @@
+"""Manage filesystems on remote FTP servers.
+"""
+
 from __future__ import print_function
 from __future__ import unicode_literals
 
@@ -10,7 +13,6 @@ import threading
 
 from contextlib import contextmanager
 from ftplib import FTP
-from ftplib import error_reply
 from ftplib import error_perm
 from ftplib import error_temp
 
@@ -26,6 +28,7 @@ from .info import Info
 from .iotools import line_iterator
 from .mode import Mode
 from .path import abspath
+from .path import dirname
 from .path import basename
 from .path import normpath
 from .path import split
@@ -44,16 +47,18 @@ def ftp_errors(fs, path=None):
         raise errors.RemoteConnectionError(
             msg='unable to connect to {}'.format(fs.host)
         )
-    except error_temp as e:
+    except error_temp as error:
         if path is not None:
             raise errors.ResourceError(
                 path,
-                msg="ftp error on resource '{}' ({})".format(path, e)
+                msg="ftp error on resource '{}' ({})".format(path, error)
             )
         else:
-            raise errors.OperationFailed(msg='ftp error ({})'.format(e))
-    except error_perm as e:
-        code, message = parse_ftp_error(e)
+            raise errors.OperationFailed(
+                msg='ftp error ({})'.format(error)
+            )
+    except error_perm as error:
+        code, message = _parse_ftp_error(error)
         if code == 552:
             raise errors.InsufficientStorage(
                 path=path,
@@ -76,20 +81,31 @@ def manage_ftp(ftp):
         except:  # pragma: nocover
             pass
 
-def parse_ftp_error(e):
-    code, _, message = text_type(e).partition(' ')
+
+def _parse_ftp_error(error):
+    """Extract code and message from ftp error."""
+    code, _, message = text_type(error).partition(' ')
     if code.isdigit():
         code = int(code)
     return code, message
 
 
-def _encode(s):
-    return s.encode() if PY2 and isinstance(s, text_type) else s
+if PY2:
+    def _encode(st, encoding):
+        return st.encode(encoding) if isinstance(st, text_type) else st
+    def _decode(st, encoding):
+        return st.decode(encoding, 'replace') if isinstance(st, bytes) else st
+else:
+    def _encode(st, _):
+        return st
+    def _decode(st, _):
+        return st
 
 
 class FTPFile(io.IOBase):
 
     def __init__(self, ftpfs, path, mode):
+        super(FTPFile, self).__init__()
         self.fs = ftpfs
         self.path = path
         self.mode = Mode(mode)
@@ -100,15 +116,16 @@ class FTPFile(io.IOBase):
         self._write_conn = None
 
     def _open_ftp(self):
+        """Open an ftp object for the file."""
         ftp = self.fs._open_ftp()
-        ftp.voidcmd(_encode('TYPE I'))
+        ftp.voidcmd(str('TYPE I'))
         return ftp
 
     @property
     def read_conn(self):
         if self._read_conn is None:
             self._read_conn = self.ftp.transfercmd(
-                _encode('RETR ' + self.path),
+                str('RETR ') + _encode(self.path, self.ftp.encoding),
                 self.pos
             )
         return self._read_conn
@@ -118,11 +135,11 @@ class FTPFile(io.IOBase):
         if self._write_conn is None:
             if self.mode.appending:
                 self._write_conn = self.ftp.transfercmd(
-                    _encode('APPE ' + self.path)
+                    str('APPE ') + _encode(self.path, self.ftp.encoding)
                 )
             else:
                 self._write_conn = self.ftp.transfercmd(
-                    _encode('STOR ' + self.path),
+                    str('STOR ') + _encode(self.path, self.ftp.encoding),
                     self.pos
                 )
         return self._write_conn
@@ -154,7 +171,7 @@ class FTPFile(io.IOBase):
     def readable(self):
         return self.mode.reading
 
-    def read(self, size=None):
+    def read(self, size=-1):
         if not self.mode.reading:
             raise IOError('File not open for reading')
 
@@ -163,8 +180,8 @@ class FTPFile(io.IOBase):
 
         conn = self.read_conn
         with self._lock:
-            while remaining is None or remaining:
-                if remaining is None:
+            while remaining:
+                if remaining < 0:
                     read_size = DEFAULT_CHUNK_SIZE
                 else:
                     read_size = min(DEFAULT_CHUNK_SIZE, remaining)
@@ -176,8 +193,7 @@ class FTPFile(io.IOBase):
                     break
                 chunks.append(chunk)
                 self.pos += len(chunk)
-                if remaining is not None:
-                    remaining -= len(chunk)
+                remaining -= len(chunk)
         return b''.join(chunks)
 
     def readline(self, size=None):
@@ -212,6 +228,8 @@ class FTPFile(io.IOBase):
                 remaining_data -= sent_size
                 self.pos += sent_size
 
+        return data_pos
+
     def writelines(self, lines):
         self.write(b''.join(lines))
 
@@ -226,6 +244,7 @@ class FTPFile(io.IOBase):
                 f.write(data)
                 if len(data) < size:
                     f.write(b'\0' * (size - len(data)))
+        return size
 
     def seekable(self):
         return True
@@ -252,23 +271,25 @@ class FTPFile(io.IOBase):
             if self._write_conn:
                 self._write_conn.close()
                 self._write_conn = None
+        return self.tell()
 
 
 class FTPFS(FS):
-    """
-    A FTP (File Transport Protocol) Filesystem.
+    """A FTP (File Transport Protocol) Filesystem.
 
-    :param str host: A FTP host, e.g. ``'ftp.mirror.nl'``.
-    :param str user: A username (default is ``'anonymous'``)
-    :param passwd: Password for the server, or ``None`` for anon.
-    :param acct: FTP account.
-    :param int timeout: Timeout for contacting server (in seconds).
-    :param int port: Port number (default 21).
+    Arguments:
+        host (str): A FTP host, e.g. ``'ftp.mirror.nl'``.
+        user (str, optional): A username (default is ``'anonymous'``).
+        passwd (str, optional): Password for the server, or `None` for anon.
+        acct (str, optional): FTP account.
+        timeout (int, optional): Timeout for contacting server (in seconds,
+            defaults to 10).
+        port (int, optional): FTP port number (default 21).
+        proxy (str): An FTP proxy, or ``None`` (default) for no proxy.
 
     """
 
     _meta = {
-        'case_insensitive': False,
         'invalid_path_chars': '\0',
         'network': True,
         'read_only': False,
@@ -283,15 +304,18 @@ class FTPFS(FS):
                  passwd='',
                  acct='',
                  timeout=10,
-                 port=21):
+                 port=21,
+                 proxy=None):
         super(FTPFS, self).__init__()
-        self.host = host
-        self.user = user
+        self._host = host
+        self._user = user
         self.passwd = passwd
         self.acct = acct
         self.timeout = timeout
         self.port = port
+        self.proxy = proxy
 
+        self.encoding = 'latin-1'
         self._ftp = None
         self._welcome = None
         self._features = None
@@ -307,12 +331,59 @@ class FTPFS(FS):
         )
         return _fmt.format(host=self.host, port=self.port)
 
+    @property
+    def user(self):
+        return (
+            self._user
+            if self.proxy is None else
+            '{}@{}'.format(self._user, self._host)
+        )
+
+    @property
+    def host(self):
+        return (
+            self._host
+            if self.proxy is None else
+            self.proxy
+        )
+
+    @classmethod
+    def _parse_features(cls, feat_response):
+        """Parse a dict of features from FTP feat response."""
+        features = {}
+        if feat_response.split('-')[0] == '211':
+            for line in feat_response.splitlines():
+                if line.startswith(' '):
+                    key, _, value = line[1:].partition(' ')
+                    features[key] = value
+        return features
+
     def _open_ftp(self):
+        """Open a new ftp object."""
         _ftp = FTP()
         _ftp.set_debuglevel(0)
         with ftp_errors(self):
             _ftp.connect(self.host, self.port, self.timeout)
             _ftp.login(self.user, self.passwd, self.acct)
+            self._features = {}
+            try:
+                feat_response = _decode(_ftp.sendcmd("FEAT"), 'latin-1')
+            except error_perm:
+                self.encoding = 'latin-1'
+            else:
+                self._features = self._parse_features(feat_response)
+                self.encoding = (
+                    'utf-8'
+                    if 'UTF8' in self._features
+                    else 'latin-1'
+                )
+                if not PY2:
+                    _ftp.file = _ftp.sock.makefile(
+                        'r',
+                        encoding=self.encoding
+                    )
+        _ftp.encoding = self.encoding
+        self._welcome = _ftp.welcome
         return _ftp
 
     def _manage_ftp(self):
@@ -321,6 +392,7 @@ class FTPFS(FS):
 
     @property
     def ftp_url(self):
+        """Get the FTP url this filesystem will open."""
         url = (
             "ftp://{}".format(self.host)
             if self.port == 21
@@ -330,28 +402,20 @@ class FTPFS(FS):
 
     @property
     def ftp(self):
+        """`~ftplib.FTP`: the underlying FTP client.
+        """
+        return self._get_ftp()
+
+    def _get_ftp(self):
         if self._ftp is None:
             self._ftp = self._open_ftp()
-            self._welcome = self._ftp.getwelcome()
         return self._ftp
 
     @property
     def features(self):
-        """Get features dict from FTP server."""
-        if self._features is None:
-            try:
-                response = self.ftp.sendcmd(_encode("FEAT"))
-            except error_perm:
-                self._features = {}
-            else:
-                self._features = {}
-                if PY2:
-                    response = response.decode('ascii')
-                if response.split('-')[0] == '211':
-                    for line in response.splitlines():
-                        if line.startswith(' '):
-                            k, _, v = line[1:].partition(' ')
-                            self._features[k] = v
+        """`dict`: features of the remote FTP server.
+        """
+        self._get_ftp()
         return self._features
 
     def _read_dir(self, path):
@@ -359,7 +423,7 @@ class FTPFS(FS):
         lines = []
         with ftp_errors(self, path=path):
             self.ftp.retrlines(
-                _encode('LIST {}'.format(_path)),
+                str('LIST ') + _encode(_path, self.ftp.encoding),
                 lines.append
             )
         lines = [
@@ -372,28 +436,29 @@ class FTPFS(FS):
 
     @property
     def supports_mlst(self):
+        """Check if server supports MLST feature."""
         return 'MLST' in self.features
 
     def create(self, path, wipe=False):
-        self.check()
-        self.validatepath(path)
+        _path = self.validatepath(path)
         with ftp_errors(self, path):
             if wipe or not self.isfile(path):
                 empty_file = io.BytesIO()
                 self.ftp.storbinary(
-                    _encode("STOR {}".format(path)),
+                    str("STOR ") + _encode(_path, self.ftp.encoding),
                     empty_file
                 )
 
     @classmethod
-    def _parse_ftp_time(cls, t):
+    def _parse_ftp_time(cls, time_text):
+        """Parse a time from an ftp directory listing."""
         try:
-            tm_year = int(t[0:4])
-            tm_month = int(t[4:6])
-            tm_day = int(t[6:8])
-            tm_hour = int(t[8:10])
-            tm_min = int(t[10:12])
-            tm_sec = int(t[12:14])
+            tm_year = int(time_text[0:4])
+            tm_month = int(time_text[4:6])
+            tm_day = int(time_text[6:8])
+            tm_hour = int(time_text[8:10])
+            tm_min = int(time_text[10:12])
+            tm_sec = int(time_text[12:14])
         except ValueError:
             return None
         epoch_time = calendar.timegm((
@@ -411,11 +476,11 @@ class FTPFS(FS):
         name = None
         facts = {}
         for fact in line.split(';'):
-            k, sep, v = fact.partition('=')
+            key, sep, value = fact.partition('=')
             if sep:
-                k = k.strip().lower()
-                v = v.strip()
-                facts[k] = v
+                key = key.strip().lower()
+                value = value.strip()
+                facts[key] = value
             else:
                 name = basename(fact.rstrip('/').strip())
         return name if name not in ('.', '..') else None, facts
@@ -456,7 +521,6 @@ class FTPFS(FS):
             yield raw_info
 
     def getinfo(self, path, namespaces=None):
-        self.check()
         _path = self.validatepath(path)
         namespaces = namespaces or ()
 
@@ -477,11 +541,9 @@ class FTPFS(FS):
             with self._lock:
                 with ftp_errors(self, path=path):
                     response = self.ftp.sendcmd(
-                        _encode('MLST {}'.format(_path))
+                        str('MLST ') + _encode(_path, self.ftp.encoding)
                     )
-                if PY2:
-                    response = response.decode('utf-8')
-                lines = response.splitlines()[1:-1]
+                lines = _decode(response, self.ftp.encoding).splitlines()[1:-1]
                 for raw_info in self._parse_mlsx(lines):
                     return Info(raw_info)
 
@@ -493,8 +555,16 @@ class FTPFS(FS):
             info = directory[file_name]
             return info
 
+    def getmeta(self, namespace="standard"):
+        _meta = {}
+        self._get_ftp()
+        if namespace == "standard":
+            _meta = self._meta.copy()
+            _meta['unicode_paths'] = "UTF8" in self.features
+        return _meta
+
+
     def listdir(self, path):
-        self.check()
         _path = self.validatepath(path)
         with self._lock:
             dir_list = [
@@ -504,9 +574,7 @@ class FTPFS(FS):
         return dir_list
 
     def makedir(self, path, permissions=None, recreate=False):
-        self.check()
-        self.validatepath(path)
-        _path = abspath(normpath(path))
+        _path = self.validatepath(path)
 
         with ftp_errors(self, path=path):
             if _path == '/':
@@ -517,9 +585,9 @@ class FTPFS(FS):
 
             if not (recreate and self.isdir(path)):
                 try:
-                    self.ftp.mkd(_encode(_path))
-                except error_perm as e:
-                    code, _ = parse_ftp_error(e)
+                    self.ftp.mkd(_encode(_path, self.ftp.encoding))
+                except error_perm as error:
+                    code, _ = _parse_ftp_error(error)
                     if code == 550:
                         if self.isdir(path):
                             raise errors.DirectoryExists(path)
@@ -532,47 +600,43 @@ class FTPFS(FS):
     def openbin(self, path, mode="r", buffering=-1, **options):
         _mode = Mode(mode)
         _mode.validate_bin()
-        self.check()
-        self.validatepath(path)
+        _path = self.validatepath(path)
 
         with self._lock:
             try:
-                info = self.getinfo(path)
+                info = self.getinfo(_path)
             except errors.ResourceNotFound:
                 if _mode.reading:
                     raise errors.ResourceNotFound(path)
+                if _mode.writing and not self.isdir(dirname(_path)):
+                    raise errors.ResourceNotFound(path)
             else:
                 if info.is_dir:
                     raise errors.FileExpected(path)
             if _mode.exclusive:
                 raise errors.FileExists(path)
-            f = FTPFile(self, abspath(normpath(path)), mode)
-        return f
+            ftp_file = FTPFile(self, _path, mode)
+        return ftp_file
 
     def remove(self, path):
         self.check()
-        self.validatepath(path)
-        _path = abspath(normpath(path))
-        dir_name, file_name = split(_path)
+        _path = self.validatepath(path)
         with self._lock:
             if self.isdir(path):
                 raise errors.FileExpected(path=path)
             with ftp_errors(self, path):
-                self.ftp.delete(_encode(_path))
+                self.ftp.delete(_encode(_path, self.ftp.encoding))
 
     def removedir(self, path):
-        self.check()
-        _path = abspath(normpath(path))
-        self.validatepath(path)
+        _path = self.validatepath(path)
         if _path == '/':
             raise errors.RemoveRootError()
-        _dir_name, file_name = split(_path)
 
         with ftp_errors(self, path):
             try:
-                self.ftp.rmd(_encode(_path))
-            except error_perm as e:
-                code, _ = parse_ftp_error(e)
+                self.ftp.rmd(_encode(_path, self.ftp.encoding))
+            except error_perm as error:
+                code, _ = _parse_ftp_error(error)
                 if code == 550:
                     if self.isfile(path):
                         raise errors.DirectoryExpected(path)
@@ -581,18 +645,17 @@ class FTPFS(FS):
                 raise  # pragma: no cover
 
     def _scandir(self, path, namespaces=None):
-        self.check()
-        self.validatepath(path)
-        _path = abspath(normpath(path))
+        _path = self.validatepath(path)
         with self._lock:
             if self.supports_mlst:
                 lines = []
                 with ftp_errors(self, path=path):
                     try:
                         self.ftp.retrlines(
-                            _encode("MLSD {}".format(_path)), lines.append
+                            str("MLSD ") + _encode(_path, self.ftp.encoding),
+                            lambda l: lines.append(_decode(l, self.ftp.encoding))
                         )
-                    except error_perm as e:
+                    except error_perm:
                         if not self.getinfo(path).is_dir:
                             raise errors.DirectoryExpected(path)
                         raise # pragma: no cover
@@ -614,43 +677,36 @@ class FTPFS(FS):
         return iter_info
 
     def setbinfile(self, path, file):
-        _path = abspath(normpath(path))
-        self.validatepath(path)
+        _path = self.validatepath(path)
         with self._lock:
             with self._manage_ftp() as ftp:
                 with ftp_errors(self, path):
                     ftp.storbinary(
-                        "STOR {}".format(_encode(_path)),
+                        str("STOR ") + _encode(_path, self.ftp.encoding),
                         file
                     )
 
     def setbytes(self, path, contents):
         if not isinstance(contents, bytes):
-            raise ValueError('contents must be bytes')
-        _path = abspath(normpath(path))
-        self.validatepath(path)
-        bin_file = io.BytesIO(contents)
-        with self._lock:
-            with self._manage_ftp() as ftp:
-                with ftp_errors(self, path):
-                    ftp.storbinary(
-                        "STOR {}".format(_encode(_path)),
-                        bin_file
-                    )
+            raise TypeError('contents must be bytes')
+        self.setbinfile(path, io.BytesIO(contents))
 
     def setinfo(self, path, info):
         if not self.exists(path):
             raise errors.ResourceNotFound(path)
 
     def getbytes(self, path):
-        _path = abspath(normpath(path))
+        _path = self.validatepath(path)
         data = io.BytesIO()
         with ftp_errors(self, path):
             with self._manage_ftp() as ftp:
                 try:
-                    ftp.retrbinary("RETR {}".format(_path), data.write)
-                except error_perm as e:
-                    code, _ = parse_ftp_error(e)
+                    ftp.retrbinary(
+                        str("RETR ") + _encode(_path, self.ftp.encoding),
+                        data.write
+                    )
+                except error_perm as error:
+                    code, _ = _parse_ftp_error(error)
                     if code == 550:
                         if self.isdir(path):
                             raise errors.FileExpected(path)
@@ -663,12 +719,7 @@ class FTPFS(FS):
         if not self.isclosed():
             try:
                 self.ftp.quit()
-            except Exception:
+            except Exception:  # pragma: no cover
                 pass
             self._ftp = None
         super(FTPFS, self).close()
-
-
-if __name__ == "__main__":  # pragma: no cover
-    fs = FTPFS('ftp.mirror.nl', 'anonymous', 'willmcgugan@gmail.com')
-    print(list(fs.scandir('/')))
