@@ -7,8 +7,7 @@ searching etc. See :ref:`walking` for details.
 
 from __future__ import unicode_literals
 
-import itertools
-
+from collections import defaultdict
 from collections import deque
 from collections import namedtuple
 
@@ -26,86 +25,7 @@ Step = namedtuple('Step', 'path, dirs, files')
 """
 
 
-class WalkerBase(object):
-    """The base class for a Walker.
-
-    To create a custom walker, implement `~fs.walk.WalkerBase.walk`
-    in a sub-class.
-
-    See `~fs.walk.Walker` for a fully featured walker object that
-    should be adequate for all but your most exotic directory walking
-    needs.
-
-    """
-
-    def walk(self, fs, path='/', namespaces=None):
-        """Walk the directory structure of a filesystem.
-
-        Arguments:
-            fs (FS): A filesystem instance.
-            path (str, optional): A path to a directory on the filesystem.
-            namespaces(list, optional): A list of additional namespaces
-                to add to the `~fs.info.Info` objects.
-
-        Returns:
-            ~collections.Iterator: iterator of `~fs.walk.Step` named tuples.
-
-        """
-        raise NotImplementedError
-
-    def files(self, fs, path='/'):
-        """Walk a filesystem, yielding absolute paths to files.
-
-        Arguments:
-            fs (FS): A filesystem instance.
-            path (str, optional): A path to a directory on the filesystem.
-
-        Yields:
-            str: absolute path to files on the filesystem found
-            recursively within the given directory.
-
-        """
-        iter_walk = iter(self.walk(fs, path=path))
-        for _path, _, files in iter_walk:
-            for info in files:
-                yield join(_path, info.name)
-
-    def dirs(self, fs, path='/'):
-        """Walk a filesystem, yielding absolute paths to directories.
-
-        Arguments:
-            fs (FS): A filesystem instance.
-            path (str, optional): A path to a directory on the filesystem.
-
-        Yields:
-            str: absolute path to directories on the filesystem found
-            recursively within the given directory.
-
-        """
-        for path, dirs, _ in self.walk(fs, path=path):
-            for info in dirs:
-                yield join(path, info.name)
-
-    def info(self, fs, path='/', namespaces=None):
-        """Walk a filesystem, yielding tuples of ``(<path>, <info>)``.
-
-        Arguments:
-            fs (FS): A filesystem instance.
-            path (str, optional): A path to a directory on the filesystem.
-            namespaces (list, optional): A list of additional namespaces
-                to add to the `Info` objects.
-
-        Yields:
-            (str, Info): a tuple of ``(<absolute path>, <resource info>)``.
-
-        """
-        _walk = self.walk(fs, path=path, namespaces=namespaces)
-        for _path, dirs, files in _walk:
-            for info in itertools.chain(dirs, files):
-                yield join(_path, info.name), info
-
-
-class Walker(WalkerBase):
+class Walker(object):
     """A walker object recursively lists directories in a filesystem.
 
     Arguments:
@@ -122,7 +42,8 @@ class Walker(WalkerBase):
             list of filename patterns, e.g. ``['*.py']``. Files will only
             be returned if the final component matches one of the patterns.
         exclude_dirs (list, optional): A list of patterns that will be used
-            to filter out directories from the walk. e.g. ``['*.svn', '*.git']``.
+            to filter out directories from the walk. e.g.
+            ``['*.svn', '*.git']``.
         max_depth (int, optional): Maximum directory depth to walk.
 
     """
@@ -223,32 +144,18 @@ class Walker(WalkerBase):
             max_depth=(self.max_depth, None)
         )
 
-    def filter_files(self, fs, infos):
-        """Filter a sequence of resource `Info` objects.
-
-        The default implementation filters those files for which
-        `~fs.walk.Walker.check_file` returns `True`.
-
-        Arguments:
-            fs (FS): a filesystem instance.
-            infos (list): A list of `~fs.info.Info` instances.
-
-        Returns:
-            list: a list of `Info` objects passing the ``check_file``
-            validation.
-
-        """
-        _check_file = self.check_file
-        return [
-            info for info in infos
-            if _check_file(fs, info)
-        ]
+    def _iter_walk(self, fs, path, namespaces=None):
+        """Get the walk generator."""
+        if self.search == 'breadth':
+            return self._walk_breadth(fs, path, namespaces=namespaces)
+        else:
+            return self._walk_depth(fs, path, namespaces=namespaces)
 
     def _check_open_dir(self, fs, path, info):
         """Check if a directory should be considered in the walk.
         """
         if (self.exclude_dirs is not None and
-            fs.match(self.exclude_dirs, info.name)):
+                fs.match(self.exclude_dirs, info.name)):
             return False
         return self.check_open_dir(fs, path, info)
 
@@ -357,13 +264,68 @@ class Walker(WalkerBase):
 
         """
         _path = abspath(normpath(path))
+        dir_info = defaultdict(list)
+        _walk = self._iter_walk(fs, _path, namespaces=namespaces)
+        for dir_path, info in _walk:
+            if info is None:
+                dirs = []
+                files = []
+                for _info in dir_info[dir_path]:
+                    (dirs if _info.is_dir else files).append(_info)
+                yield Step(dir_path, dirs, files)
+                del dir_info[dir_path]
+            else:
+                dir_info[dir_path].append(info)
 
-        if self.search == 'breadth':
-            do_walk = self._walk_breadth
-        elif self.search == 'depth':
-            do_walk = self._walk_depth
+    def files(self, fs, path='/'):
+        """Walk a filesystem, yielding absolute paths to files.
 
-        return do_walk(fs, _path, namespaces=namespaces)
+        Arguments:
+            fs (FS): A filesystem instance.
+            path (str, optional): A path to a directory on the filesystem.
+
+        Yields:
+            str: absolute path to files on the filesystem found
+            recursively within the given directory.
+
+        """
+        for _path, info in self._iter_walk(fs, path=path):
+            if info is not None and not info.is_dir:
+                yield join(_path, info.name)
+
+    def dirs(self, fs, path='/'):
+        """Walk a filesystem, yielding absolute paths to directories.
+
+        Arguments:
+            fs (FS): A filesystem instance.
+            path (str, optional): A path to a directory on the filesystem.
+
+        Yields:
+            str: absolute path to directories on the filesystem found
+            recursively within the given directory.
+
+        """
+        for _path, info in self._iter_walk(fs, path=path):
+            if info is not None and info.is_dir:
+                yield join(_path, info.name)
+
+    def info(self, fs, path='/', namespaces=None):
+        """Walk a filesystem, yielding tuples of ``(<path>, <info>)``.
+
+        Arguments:
+            fs (FS): A filesystem instance.
+            path (str, optional): A path to a directory on the filesystem.
+            namespaces (list, optional): A list of additional namespaces
+                to add to the `Info` objects.
+
+        Yields:
+            (str, Info): a tuple of ``(<absolute path>, <resource info>)``.
+
+        """
+        _walk = self._iter_walk(fs, path=path, namespaces=namespaces)
+        for _path, info in _walk:
+            if info is not None:
+                yield join(_path, info.name), info
 
     def _walk_breadth(self, fs, path, namespaces=None):
         """Walk files using a *breadth first* search.
@@ -375,22 +337,17 @@ class Walker(WalkerBase):
 
         while queue:
             dir_path = pop()
-            dirs = []
-            files = []
             for info in self._scan(fs, dir_path, namespaces=namespaces):
                 if info.is_dir:
                     _depth = self._calculate_depth(dir_path) - depth + 1
                     if self._check_open_dir(fs, dir_path, info):
-                        dirs.append(info)
+                        yield dir_path, info  # Opened a directory
                         if self._check_scan_dir(fs, dir_path, info, _depth):
                             push(join(dir_path, info.name))
                 else:
-                    files.append(info)
-            yield Step(
-                dir_path,
-                dirs,
-                self.filter_files(fs, files)
-            )
+                    if self.check_file(fs, info):
+                        yield dir_path, info  # Found a file
+            yield dir_path, None  # End of directory
 
     def _walk_depth(self, fs, path, namespaces=None):
         """Walk files using a *depth first* search.
@@ -402,34 +359,28 @@ class Walker(WalkerBase):
             return self._scan(fs, path, namespaces=namespaces)
 
         depth = self._calculate_depth(path)
-        stack = [(
-            path, scan(path), [], []
-        )]
+        stack = [(path, scan(path), None)]
         push = stack.append
 
         while stack:
-            dir_path, iter_files, dirs, files = stack[-1]
-            try:
-                info = next(iter_files)
-            except StopIteration:
-                yield Step(
-                    dir_path,
-                    dirs,
-                    self.filter_files(fs, files)
-                )
+            dir_path, iter_files, parent = stack[-1]
+            info = next(iter_files, None)
+            if info is None:
+                if parent is not None:
+                    yield parent
+                yield dir_path, None
                 del stack[-1]
+            elif info.is_dir:
+                _depth = self._calculate_depth(dir_path) - depth + 1
+                if self._check_open_dir(fs, dir_path, info):
+                    if self._check_scan_dir(fs, dir_path, info, _depth):
+                        _path = join(dir_path, info.name)
+                        push((_path, scan(_path), (dir_path, info)))
+                    else:
+                        yield dir_path, info
             else:
-                if info.is_dir:
-                    _depth = self._calculate_depth(dir_path) - depth + 1
-                    if self._check_open_dir(fs, dir_path, info):
-                        dirs.append(info)
-                        if self._check_scan_dir(fs, dir_path, info, _depth):
-                            _path = join(dir_path, info.name)
-                            push((
-                                _path, scan(_path), [], []
-                            ))
-                else:
-                    files.append(info)
+                if self.check_file(fs, info):
+                    yield dir_path, info
 
 
 class BoundWalker(object):
